from typing_extensions import TypeVar, Generic, Callable, Awaitable, Any, Protocol, overload, Coroutine
from dataclasses import dataclass
import asyncio
from datetime import timedelta, datetime
import traceback
from sqlmodel import SQLModel, Field, select, Session, or_
from sqlalchemy import delete
from pydantic import TypeAdapter
from sqltypes import ValidatedJSON
from pipeteer.pipelines import Pipeline, Inputtable, Context, Input, Entry, InputT, EntryT
from pipeteer.util import param_type, return_type, race
from .pipeline import A

Aw = Awaitable
# A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
AnyT: type = Any # type: ignore

class Stop(BaseException):
  ...

class WorkflowContext(Protocol):
  async def call(self, pipe: Inputtable[A, B], x: A, /) -> B:
    ...
  @overload
  async def all(self, a: Aw[A], b: Aw[B], /) -> tuple[A, B]: ...
  @overload
  async def all(self, a: Aw[A], b: Aw[B], c: Aw[C], /) -> tuple[A, B, C]: ...
  @overload
  async def all(self, a: Aw[A], b: Aw[B], c: Aw[C], d: Aw[D], /) -> tuple[A, B, C, D]: ...
  @overload
  async def all(self, *coros: Aw[A]) -> tuple[A, ...]: ...

@dataclass
class WkfContext(WorkflowContext):
  ctx: Context
  session: Session
  State: type['State']
  states: list['State']
  key: str
  output: str
  step: int = 0

  async def call(self, pipe: Inputtable[A, B], x: A, /) -> B:
    if self.step < len(self.states):
      state = self.states[self.step]
      if pipe.input_adapter.dump_python(x) != state.param:
        param = pipe.input_adapter.validate_python(state.param)
        self.ctx.log(f'Impure workflow. At step {self.step}, calling "{pipe.id}": expected "{param}" but got "{x}"', level='ERROR')
        raise RuntimeError('Impure workflow')

      self.ctx.log(f'Replaying {pipe.id}, step={self.step}, key="{self.key}"', level='DEBUG')

      self.step += 1
      return pipe.output_adapter.validate_python(state.result)
    
    else:
      self.ctx.log(f'Calling {pipe.id}, step={self.step}, key="{self.key}"', level='DEBUG')
      PipeInp = pipe.input(self.ctx)
      with Session(self.ctx.db.engine) as s:
        s.add(PipeInp(key=f'{self.step}_{self.key}', value=x, output=self.output))
        s.add(self.State(key=self.key, step=self.step, param=x, pipeline=pipe.id))
        s.commit()
      await self.ctx.zmq.pub.send(pipe.id)

      self.step += 1
      raise Stop()
    
  async def all(self, *coros: Awaitable): # type: ignore
    n = len(coros)
    if self.step + n-1 < len(self.states):
      self.ctx.log(f'Replaying all, step={self.step}, key="{self.key}"', level='DEBUG')
      return tuple(await asyncio.gather(*coros))
    
    elif self.step == len(self.states):
      self.ctx.log(f'Calling all, step={self.step}, key="{self.key}"', level='DEBUG')
      for coro in coros:
        try:
          await coro
        except Stop:
          ...
    else:
      received = self.step+n+1 - len(self.states)
      self.ctx.log(f'Ignoring all (received {received}/{n}), step={self.step}, key="{self.key}"', level='DEBUG')
    raise Stop()
  
class State(SQLModel):
  key: str = Field(primary_key=True)
  step: int = Field(primary_key=True)
  param: Any = Field(sa_type=ValidatedJSON(AnyT))
  result: Any | None = Field(default=None, sa_type=ValidatedJSON(AnyT))
  done: bool = False
  pipeline: str
  
class WkfInputT(InputT[A], Generic[A]):
  doing: bool = False

def WkfInput(T: type[A]) -> type[WkfInputT[A]]:
  Inp = Input(T)
  class WkfInput(Inp):
    doing: bool = False
  return WkfInput # type: ignore


class Result(EntryT):
  value: Any = Field(sa_type=ValidatedJSON(AnyT))
  ttl: datetime | None = None

@dataclass
class Workflow(Pipeline[A, B, Context, Coroutine], Generic[A, B]):
  Input = WkfInput
  State = State
  call: Callable[[A, WorkflowContext], Awaitable[B]]
  reserve: timedelta | None = None
  poll_interval: timedelta = timedelta(seconds=1)

  def states(self, ctx: Context):
    return ctx.db.table(self.id + '-states', State)
  
  def input(self, ctx: Context):
    return ctx.db.table(self.id, WkfInput(self.Tin))
  
  def results(self, ctx: Context):
    output = self.id + '-results'
    table = ctx.db.table(output, Result)
    return output, table
  
  def run(self, ctx: Context) -> Coroutine:
      
    self.results(ctx) # trigger creation
    
    async def loop():

      # trigger creation (in-process)
      ctx.zmq.pub
      ctx.zmq.sub(self.id)
      ctx.zmq.sub(self.id + '-results')

      output, Result = self.results(ctx)
      Input = self.input(ctx)
      State = self.states(ctx)

      async def run(*, key: str, results_key: str | None = None, input, states: list['State'] = [], s: Session):
        wkf_ctx = WkfContext(ctx, session=s, State=State, states=states, key=key, output=output)
        ctx.log(f'Rerunning "{key}"', level='DEBUG')
        input = TypeAdapter(self.Tin).validate_python(input)
        out = await self.call(input, wkf_ctx)
        
        item = s.get(Input, key)
        if item is None:
          raise ValueError(f'Input item "{key}" not found')
        
        ctx.log(f'Outputting "{key}" to "{item.output}"', level='DEBUG')
        Output = ctx.db.table(item.output, Entry(self.Tout))
        s.add(Output(key=key, value=out))
        s.exec(delete(State).where(State.key == key)) # type: ignore[call-overload]
        s.exec(delete(Input).where(Input.key == key)) # type: ignore[call-overload]
        if results_key is not None:
          s.exec(delete(Result).where(Result.key == results_key)) # type: ignore[call-overload]

        s.commit()
        await ctx.zmq.pub.send(item.output)
        
      async def input_step(inp: WkfInputT, s: Session):
        ctx.log(f'Input loop: "{inp.key}"', level='DEBUG')
        try:
          await run(key=inp.key, input=inp.value, s=s)
        except Stop:
          inp.doing = True
          s.add(inp)
          s.commit()

      async def results_step(inp: EntryT, s: Session):
        i, key = inp.key.split('_', 1)
        i = int(i)
        ctx.log(f'Results loop: "{key}", step={i}', level='DEBUG')

        if not (input := s.get(Input, key)):
          return
        
        states = s.exec(select(State).where(State.key == key)).all()
        states = sorted(states, key=lambda state: state.step)
        if not states or states[-1].done:
          return
        
        state = states[-1]
        state.done = True
        state.result = inp.value
        s.add(state)

        try:
          await run(key=key, results_key=inp.key, input=input.value, states=states, s=s)
          
        except Stop:
          s.delete(inp)
          s.commit()


      async def input_loop():
        sub = ctx.zmq.sub(self.id)
        while True:
          try:
            with Session(ctx.db.engine) as s:
              free = or_(Input.ttl == None, Input.ttl < datetime.now()) # type: ignore
              inp = s.exec(select(Input).where(Input.doing == False, free)).first()
              if inp is None:
                await race([
                  asyncio.sleep(self.poll_interval.total_seconds()),
                  sub.wait()
                ])
                continue
              elif self.reserve:
                inp.ttl = datetime.now() + self.reserve
                s.commit()
            
              await input_step(inp, s)
          except:
            ctx.log('Error in input loop', traceback.format_exc(), level='ERROR')
            await asyncio.sleep(self.poll_interval.total_seconds())


      async def results_loop():
        sub = ctx.zmq.sub(output)
        while True:
          try:
            with Session(ctx.db.engine) as s:
              inp = s.exec(select(Result)).first()
              if inp is None:
                await race([
                  asyncio.sleep(self.poll_interval.total_seconds()),
                  sub.wait()
                ])
                continue
              elif self.reserve:
                inp.ttl = datetime.now() + self.reserve
                s.commit()
            
              await results_step(inp, s)
          except:
            ctx.log('Error in results loop', traceback.format_exc(), level='ERROR')
            await asyncio.sleep(self.poll_interval.total_seconds())

      await asyncio.gather(input_loop(), results_loop())

    return loop()
  

  async def delete(self, ctx: Context, key: str):
    """Delete a workflow by key"""
    Inp = self.input(ctx)
    State = self.states(ctx)
    _, Result = self.results(ctx)
    with Session(ctx.db.engine) as s:
      s.exec(delete(State).where(State.key == key)) # type: ignore[call-overload]
      s.exec(delete(Result).where(Result.key == key)) # type: ignore[call-overload]
      s.exec(delete(Inp).where(Inp.key == key)) # type: ignore[call-overload]
      s.commit()

  async def restart(self, ctx: Context, key: str):
    """Restart a workflow by key"""
    Inp = self.input(ctx)
    State = self.states(ctx)
    _, Result = self.results(ctx)
    with Session(ctx.db.engine) as s:
      if (inp := s.get(Inp, key)):
        inp.doing = False
        inp.ttl = None
        s.add(inp)
        s.exec(delete(State).where(State.key == key)) # type: ignore[call-overload]
        s.exec(delete(Result).where(Result.key == key)) # type: ignore[call-overload]
        s.commit()
        await ctx.zmq.pub.send(self.id)

  async def step(self, ctx: Context, key: str) -> int | None:
    """At which step is the workflow?"""
    State = self.states(ctx)
    with Session(ctx.db.engine) as s:
      state = s.exec(select(State).where(State.key == key).order_by(State.step.desc())).first() # type: ignore
      if state:
        return state.step

def workflow(
  *, id: str | None = None,
  reserve: timedelta | None = timedelta(minutes=2),
  poll_interval: timedelta = timedelta(minutes=2),
):
  def decorator(fn: Callable[[A, WorkflowContext], Awaitable[B]]) -> Workflow[A, B]:
    Tin = param_type(fn)
    if Tin is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its input parameter')

    Tout = return_type(fn)
    if Tout is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its return value')
    
    return Workflow(
      Tin=Tin, Tout=Tout,
      id=id or fn.__name__,
      call=fn, # type: ignore
      reserve=reserve,
      poll_interval=poll_interval,
    ) # type: ignore	
  return decorator


from typing_extensions import TypeVar, Generic, Callable, overload, Any, Sequence, Protocol, Union
from dataclasses import dataclass
from pipeteer.pipelines import Runnable, Context

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
T = TypeVar('T')
Ctx = TypeVar('Ctx', bound=Context)
Artif1 = TypeVar('Artif1')
Artif2 = TypeVar('Artif2')

class MultiFn(Protocol, Generic[A, Ctx, T]): # type: ignore
  def __call__(self, *artifs: A, ctx: Ctx) -> T:
    ...

@dataclass
class MultiTask(Runnable[A, B, Ctx, Artif2], Generic[A, B, Ctx, Artif1, Artif2]):
  def __init__(
    self, id: str,
    pipelines: Sequence[Runnable[A, B, Ctx, Artif1]],
    merge: MultiFn[Artif1, Ctx, Artif2]
  ):
    self.id = id
    self.pipelines = pipelines
    self.merge = merge
    self.Tin = Union[*(pipe.Tin for pipe in pipelines)] # type: ignore
    self.Tout = Union[*(pipe.Tout for pipe in pipelines)] # type: ignore

  def run(self, ctx: Ctx, /) -> Artif2:
    artifs = tuple(pipe.run(ctx) for pipe in self.pipelines)
    return self.merge(*artifs, ctx=ctx) # type: ignore

class MultiFn2(Protocol, Generic[A, B, Ctx, T]): # type: ignore
  def __call__(self, a: A, b: B, /, ctx: Ctx) -> T: ...

class MultiFn3(Protocol, Generic[A, B, C, Ctx, T]): # type: ignore
  def __call__(self, a: A, b: B, c: C, /, ctx: Ctx) -> T: ...

class MultiFn4(Protocol, Generic[A, B, C, D, Ctx, T]): # type: ignore
  def __call__(self, a: A, b: B, c: C, d: D, /, ctx: Ctx) -> T: ...

@overload
def multitask( # type: ignore
  p1: Runnable[Any, Any, Any, A],
  p2: Runnable[Any, Any, Any, B], /, *,
  id: str | None = None,
) -> Callable[[MultiFn2[A, B, Ctx, T]], MultiTask[Any, Any, Ctx, Any, T]]:
  ...
@overload
def multitask(
  p1: Runnable[Any, Any, Any, A],
  p2: Runnable[Any, Any, Any, B],
  p3: Runnable[Any, Any, Any, C], /, *,
  id: str | None = None,
) -> Callable[[MultiFn3[A, B, C, Ctx, T]], MultiTask[Any, Any, Ctx, Any, T]]:
  ...
@overload
def multitask(
  p1: Runnable[Any, Any, Any, A],
  p2: Runnable[Any, Any, Any, B],
  p3: Runnable[Any, Any, Any, C],
  p4: Runnable[Any, Any, Any, D], /, *,
  id: str | None = None,
) -> Callable[[MultiFn4[A, B, C, D, Ctx, T]], MultiTask[Any, Any, Ctx, Any, T]]:
  ...
@overload
def multitask(*pipelines: Runnable[Any, Any, Any, T], id: str | None = None) -> Callable[[MultiFn[Any, Ctx, T]], MultiTask[Any, Any, Ctx, Any, T]]:
  ...

def multitask(*pipelines: Runnable[Any, Any, Any, T], id: str | None = None): # type: ignore
  def decorator(merge: MultiFn[Any, Ctx, T]):
    return MultiTask(id or merge.__name__, pipelines, merge) # type: ignore	
  return decorator



from .backend import DB, ZMQ
from .pipelines import (
  Context, Input, InputT, Entry, EntryT,
  activity, workflow, WorkflowContext, task, Push, multitask
)

__all__ = [
  'DB', 'ZMQ',
  'Context', 'Input', 'Entry', 'InputT', 'EntryT',
  'activity',
  'workflow', 'WorkflowContext',
  'task', 'Push', 'multitask',
]


from typing_extensions import Iterable, Coroutine, Any, TypeVar, Callable, Awaitable
import asyncio

A = TypeVar('A')

async def race(coros: Iterable[Coroutine[Any, Any, A]]) -> tuple[int, A]:
  """Race `coros` and return `(idx, result)` of the first completed coroutine"""
  async def enum_task(idx, coro):
    return idx, await coro
  
  tasks = [asyncio.create_task(enum_task(i, c)) for i, c in enumerate(coros)]
  done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
  for task in pending:
    task.cancel()
  return done.pop().result()


async def exp_backoff(
  fn: Callable[[], Awaitable[A|None]], *,
  t0: float = 1, base: float = 2, tmax: float | None = None
) -> A:
  """Exponential backoff polling, waiting `t(i) = t0 * base^i` seconds before the `i`-th call to `fn`
  - `tmax`: optional upper bound. If set, `t(i) = min(t0 * base^i, tmax)`
  """
  i = 0
  while True:
    if (x := await fn()):
      return x
    t = t0 * base ** i
    if tmax is not None:
      t = min(t, tmax)
    await asyncio.sleep(t)


from typing_extensions import TypeVar, Generic, Callable, Any, Awaitable, Protocol
from dataclasses import dataclass
from sqlmodel import Session
from pipeteer.pipelines import Pipeline, Context, Entry, InputT
from pipeteer.util import param_type, type_arg, num_params, Func2or3

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D', contravariant=True)
Ctx = TypeVar('Ctx', bound=Context)
Artifact = TypeVar('Artifact')

class Push(Protocol, Generic[D]):
  async def __call__(self, key: str, val: D) -> bool:
    ...

@dataclass
class Task(Pipeline[A, B, Ctx, Artifact], Generic[A, B, Ctx, Artifact]):
  call: Callable[[type[InputT[A]], Push[B], Ctx], Artifact]

  def run(self, ctx: Ctx, /):
    Input = self.input(ctx)
    async def push(key: str, val: B):
      with Session(ctx.db.engine) as s:
        if (inp := s.get(Input, key)) is None:
          return False
        else:
          out = inp.output
          Output = ctx.db.table(out, Entry(self.Tout))
          s.add(Output(key=key, value=val))
          s.delete(inp)
          s.commit()
      
      await ctx.zmq.pub.send(out)
      return True

    return self.call(Input, push, ctx)

def task(id: str | None = None):
  def decorator(fn: Func2or3[type[InputT[A]], Push[B], Ctx, Artifact]) -> Task[A, B, Ctx, Artifact]:
    Tin = type_arg(type_arg(param_type(fn, 0) or Any)) # type: ignore
    if Tin is None:
      raise TypeError(f'Task {fn.__name__} must have a type hint for its input type')
    
    Tout = type_arg(param_type(fn, 1) or Any) # type: ignore
    if Tout is None:
      raise TypeError(f'Task {fn.__name__} must have a type hint for its output type')
    
    return Task(
      id=id or fn.__name__,
      Tin=Tin, Tout=Tout,
      call=fn if num_params(fn) == 3 else (lambda Qin, Qout, _: fn(Qin, Qout)) # type: ignore
    )
      
  return decorator


import typer

main = typer.Typer()

@main.callback()
def callback():
  ...

@main.command()
def proxy(
  pub_url: str = typer.Option('tcp://*:5555', '-p', '--pub'),
  sub_url: str = typer.Option('tcp://*:5556', '-s', '--sub'),
  verbose: bool = typer.Option(False, '-v', '--verbose')
):
  import asyncio
  from dslog import Logger
  from pipeteer.backend import proxy
  log = Logger.click().prefix('[PROXY]') if verbose else Logger.empty()
  asyncio.run(proxy(pub_url=pub_url, sub_url=sub_url, log=log))



from typing_extensions import TypeVar, Generic, Callable, Awaitable, Coroutine
from dataclasses import dataclass
import asyncio
from datetime import timedelta, datetime
from multiprocessing import Process
import traceback
from sqlmodel import select, Session, or_
from pipeteer.pipelines import Pipeline, Context, Entry
from pipeteer.util import param_type, return_type, num_params, Func1or2, race

A = TypeVar('A')
B = TypeVar('B')
Ctx = TypeVar('Ctx', bound=Context)

@dataclass
class Activity(Pipeline[A, B, Ctx, Coroutine], Generic[A, B, Ctx]):
  call: Callable[[A, Ctx], Awaitable[B]]
  reserve: timedelta | None = None
  poll_interval: timedelta = timedelta(seconds=1)

  def run(self, ctx: Ctx) -> Coroutine:
    async def loop():
      ctx.log('Running...', level='DEBUG')
      pub = ctx.zmq.pub
      sub = ctx.zmq.sub(self.id)
      Input = self.input(ctx)
      while True:
        try:
          with Session(ctx.db.engine) as s:
            free = or_(Input.ttl == None, Input.ttl < datetime.now()) # type: ignore
            if not (inp := s.exec(select(Input).where(free)).first()):
              await race([
                asyncio.sleep(self.poll_interval.total_seconds()),
                sub.wait()
              ])
              continue
            elif self.reserve:
              inp.ttl = datetime.now() + self.reserve
              s.commit()

            key, val, out = inp.key, inp.value, inp.output
          
          ctx.log(f'Processing "{key}"', level='DEBUG')

          try:
            y = await self.call(val, ctx)
            Output = ctx.db.table(out, Entry(self.Tout))

            with Session(ctx.db.engine) as s:
              s.add(Output(key=key, value=y))
              s.delete(inp)
              s.commit()

            await pub.send(out)

          except Exception:
            ctx.log(f'Error processing "{key}": {traceback.format_exc()}. Value: {val}', level='ERROR')

        except Exception:
          ctx.log(f'Error reading from input queue: {traceback.format_exc()}', level='ERROR')
      
    return loop()
  
def activity(
  id: str | None = None, *,
  reserve: timedelta | None = timedelta(minutes=2),
  poll_interval: timedelta = timedelta(minutes=2)
):
  def decorator(fn: Func1or2[A, Ctx, Awaitable[B]]) -> Activity[A, B, Ctx]:
    Tin = param_type(fn)
    if Tin is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its input parameter')

    Tout = return_type(fn)
    if Tout is None:
      raise TypeError(f'Activity {fn.__name__} must have a type hint for its return value')

    return Activity(
      Tin=Tin or param_type(fn), Tout=Tout or return_type(fn), reserve=reserve, id=id or fn.__name__,
      call=fn if num_params(fn) == 2 else (lambda x, _: fn(x)), # type: ignore
      poll_interval=poll_interval
    )
      
  return decorator


from .pipeline import Pipeline, Inputtable, Runnable, Context, Input, Entry, InputT, EntryT
from ._activity import activity, Activity
from ._workflow import workflow, Workflow, WorkflowContext
from ._task import task, Task, Push
from ._multitask import multitask, MultiTask

__all__ = [
  'Pipeline', 'Inputtable', 'Runnable', 'Context', 'Input', 'Entry', 'InputT', 'EntryT',
  'activity', 'Activity',
  'workflow', 'Workflow', 'WorkflowContext',
  'task', 'Task', 'Push',
  'multitask', 'MultiTask',
]


from typing_extensions import TypeVar
from dataclasses import dataclass, field, replace
from sqlalchemy import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.sql.schema import MetaData
from sqlmodel import SQLModel, create_engine, Session

T = TypeVar('T', bound=SQLModel)

@dataclass
class DB:
  engine: Engine
  metadata: MetaData = MetaData()
  tables: dict[str, type] = field(default_factory=dict)
  prefix_: str = ''

  def prefix(self, prefix: str, /):
    return replace(self, prefix_=prefix)

  @classmethod
  def of(cls, db_url: str, /):
    return cls(create_engine(db_url))

  @classmethod
  def at(cls, sqlite_path: str, /):
    import os
    dir = os.path.dirname(sqlite_path)
    if dir:
      os.makedirs(dir, exist_ok=True)
    return cls.of(f'sqlite:///{sqlite_path}')
  
  @property
  def session(self):
    return Session(self.engine)

  def table(self, name: str, type: type[T], /) -> type[T]:
    name = self.prefix_ + name
    # if name in self.tables:
    #   return self.tables[name]
    class _Cls(type, table=True):
      __tablename__ = name # type: ignore
    _Cls.__name__ = type.__name__
    table = _Cls.metadata.tables[name]
    _Cls.metadata._remove_table(name, table.schema)
    self.metadata._add_table(name, table.schema, table)
    try:
      self.metadata.create_all(self.engine, tables=[table])
    except OperationalError:
      ...
    self.metadata._remove_table(name, table.schema)
    self.tables[name] = _Cls

    return _Cls # type: ignore


from dataclasses import dataclass, field
import os
import zmq
from zmq.asyncio import Context
from dslog import Logger

@dataclass
class ZMQ:
  pub_url: str = 'tcp://localhost:5555'
  sub_url: str = 'tcp://localhost:5556'
  _pid: int = os.getpid()
  _subs: dict[str, 'Sub'] = field(default_factory=dict)
  _pub: 'Pub | None' = None

  def __post_init__(self):
    self._pub = Pub(self.pub_url)

  def sub(self, topic: str):
    pid = os.getpid()
    if self._pid != pid:
      self._pid = pid
      self._subs = {}

    if not topic in self._subs:
      self._subs[topic] = Sub(topic, self.sub_url)

    return self._subs[topic]

  @property  
  def pub(self):
    pid = os.getpid()
    if self._pid != pid:
      self._pid = pid
      self._pub = None

    if self._pub is None:
      self._pub = Pub(self.pub_url)

    return self._pub
  
  async def proxy(self, log: Logger = Logger.empty()):
    await proxy(pub_url=self.pub_url, sub_url=self.sub_url, log=log)

@dataclass
class Sub:
  topic: str
  url: str = 'tcp://localhost:5556'

  def __post_init__(self):
    self.pid = os.getpid()
    ctx = Context.instance()
    self.sub = ctx.socket(zmq.SUB)
    self.sub.connect(self.url)
    self.sub.setsockopt_string(zmq.SUBSCRIBE, self.topic)

  async def wait(self):
    if os.getpid() != self.pid:
      raise RuntimeError('Sub is not fork-safe')
    while True:
      topic = await self.sub.recv_string()
      if topic == self.topic: # could be a prefix but not equal
        return

@dataclass
class Pub:
  url: str = 'tcp://localhost:5555'

  def __post_init__(self):
    self.pid = os.getpid()
    ctx = Context.instance()
    self.pub = ctx.socket(zmq.PUB)
    self.pub.connect(self.url)

  async def send(self, topic: str):
    if os.getpid() != self.pid:
      raise RuntimeError('Pub is not fork-safe')
    await self.pub.send_string(topic)


async def proxy(
  pub_url: str = 'tcp://*:5555', sub_url: str = 'tcp://*:5556',
  log: Logger = Logger.empty()
):
  ctx = Context.instance()

  print(f'Proxying {pub_url} -> {sub_url} [VERBOSE]')
  frontend = ctx.socket(zmq.SUB)
  frontend.bind(pub_url)
  frontend.setsockopt_string(zmq.SUBSCRIBE, '')
  
  backend = ctx.socket(zmq.PUB)
  backend.bind(sub_url)

  while True:
    msg = await frontend.recv()
    log('Proxying:', msg, level='DEBUG')
    await backend.send(msg)





from typing_extensions import TypeVar, Generic, Self, Any, AsyncIterable
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, replace, KW_ONLY
from datetime import datetime
from functools import cached_property
from sqlmodel import SQLModel, Field, select
from pydantic import TypeAdapter
from sqltypes import ValidatedJSON
from dslog import Logger
from pipeteer.backend import DB, ZMQ

AnyT: type = Any # type: ignore
A = TypeVar('A')
B = TypeVar('B')
Artifact = TypeVar('Artifact', covariant=True)

@dataclass
class Context:
  db: DB
  zmq: ZMQ
  _: KW_ONLY
  log: Logger = field(default_factory=Logger.click)

  async def wait(self, topic: str, /):
    await self.zmq.sub(topic).wait()

  @classmethod
  def of(
    cls, db: DB, *, log: Logger = Logger.click(),
    pub_url: str = 'tcp://localhost:5555',
    sub_url: str = 'tcp://localhost:5556',
  ):
    return cls(db, ZMQ(pub_url=pub_url, sub_url=sub_url), log=log)

  def prefix(self, prefix: str) -> Self:
    return replace(self, log=self.log.prefix(prefix))
  
Ctx = TypeVar('Ctx', bound=Context)


class EntryT(SQLModel, Generic[A]):
  key: str = Field(primary_key=True)
  value: A
  
  def __class_getitem__(cls, type) -> type: # apparently SQLModel messes generics up
    cls = super().__class_getitem__(type)
    cls.__args__ = (type,) # type: ignore
    return cls # type: ignore

def Entry(type: type[A]) -> type[EntryT[A]]:
  class Entry(EntryT[type]):
    value: type = Field(sa_type=ValidatedJSON(type, name='JSON'))
  return Entry

class InputT(EntryT[A], Generic[A]):
  ttl: datetime | None = None
  output: str = 'output'


def Input(type: type[A]) -> type[InputT[A]]:
  class Input(InputT[type]):
    value: type = Field(sa_type=ValidatedJSON(type, name='JSON'))
  return Input

@dataclass
class TypesMixin(Generic[A, B]):
  Tin: type[A]
  Tout: type[B]

  @cached_property
  def input_adapter(self):
    return TypeAdapter(self.Tin)
  
  @cached_property
  def output_adapter(self):
    return TypeAdapter(self.Tout)

@dataclass
class Inputtable(TypesMixin[A, B], Generic[A, B]):
  id: str

  def input(self, ctx: Context) -> type[InputT[A]]:
    return ctx.db.table(self.id, Input(self.Tin)) # type: ignore
  
  def output(self, ctx: Context, table: str = 'output') -> type[EntryT[B]]:
    return ctx.db.table(table, Entry(self.Tout))
  
  async def notify(self, ctx: Context):
    await ctx.zmq.pub.send(self.id)

  async def items(self, ctx: Context) -> AsyncIterable[InputT[A]]:
    Inp = self.input(ctx)
    with ctx.db.session as s:
      for it in s.exec(select(Inp)):
        yield it
  

@dataclass
class Runnable(ABC, TypesMixin[A, B], Generic[A, B, Ctx, Artifact]):
  id: str
  
  @abstractmethod
  def run(self, ctx: Ctx, /) -> Artifact:
    ...

class Pipeline(Runnable[A, B, Ctx, Artifact], Inputtable[A, B], Generic[A, B, Ctx, Artifact]):
  ...


from typing_extensions import TypeVar, Callable, get_args
import inspect

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')

def param_type(fn: Callable, idx=0) -> type | None:
  sig = inspect.signature(fn)
  ann = sig.parameters[list(sig.parameters.keys())[idx]].annotation
  if ann != inspect.Signature.empty:
    return ann

def return_type(fn: Callable) -> type | None:
  ann = inspect.signature(fn).return_annotation
  if ann != inspect.Signature.empty:
    return ann


def type_arg(generic: type, idx=0) -> type | None:
  try:
    return get_args(generic)[idx]
  except IndexError:
    ...
  try:
    return generic.__args__[idx]
  except:
    ...

def num_params(fn) -> int:
  return len(inspect.signature(fn).parameters)

Func1or2 = Callable[[A, B], C] | Callable[[A], C]
Func2or3 = Callable[[A, B, C], D] | Callable[[A, B], D]